--- cfi_cmdset_0002.c	Fri Jul 21 10:30:50 2000
+++ /home/simon/mtd/kernel/cfi_cmdset_0002.c	Mon Jul 24 14:26:09 2000
@@ -4,6 +4,8 @@
  *
  * Copyright (C) 2000 Crossnet Co. <info@crossnet.co.jp>
  *
+ * 2_by_8 routines added by Simon Munton
+ *
  * This code is GPL
  *
  * $Id: cfi_cmdset_0002.c,v 1.1 2000/07/11 12:32:09 dwmw2 Exp $
@@ -30,6 +32,9 @@
 static int cfi_amdext_read_2_by_16 (struct mtd_info *, loff_t, size_t, size_t *, u_char *);
 static int cfi_amdext_write_2_by_16(struct mtd_info *, loff_t, size_t, size_t *, const u_char *);
 static int cfi_amdext_erase_2_by_16 (struct mtd_info *, struct erase_info *);
+static int cfi_amdext_read_2_by_8 (struct mtd_info *, loff_t, size_t, size_t *, u_char *);
+static int cfi_amdext_write_2_by_8 (struct mtd_info *, loff_t, size_t, size_t *, const u_char *);
+static int cfi_amdext_erase_2_by_8 (struct mtd_info *, struct erase_info *);
 static void cfi_amdext_sync (struct mtd_info *);
 static int cfi_amdext_suspend (struct mtd_info *);
 static void cfi_amdext_resume (struct mtd_info *);
@@ -93,9 +98,57 @@
 	mtd->erasesize = 0x20000; /* FIXME */
 	/* Also select the correct geometry setup too */ 
 	mtd->size = (1 << cfi->cfiq.DevSize) * cfi->numchips * cfi->interleave;
-	mtd->erase = cfi_amdext_erase_2_by_16;
-	mtd->read = cfi_amdext_read_2_by_16;
-	mtd->write = cfi_amdext_write_2_by_16;
+
+	switch (map->buswidth)
+	{
+	case 1:
+	  printk("Unsupported buswidth\n");
+		kfree(mtd);
+	  kfree(cfi->cmdset_priv);
+	  return NULL;
+		break;
+
+	case 2:
+		switch (cfi->interleave)
+		{
+		case 2:
+			mtd->erase = cfi_amdext_erase_2_by_8;
+			mtd->read = cfi_amdext_read_2_by_8;
+			mtd->write = cfi_amdext_write_2_by_8;
+			break;
+
+		default:
+		  printk("Unsupported interleave\n");
+			kfree(mtd);
+	  	kfree(cfi->cmdset_priv);
+		  return NULL;
+		}
+    break;
+
+	case 4:
+		switch (cfi->interleave)
+		{
+		case 2:
+			mtd->erase = cfi_amdext_erase_2_by_16;
+			mtd->read = cfi_amdext_read_2_by_16;
+			mtd->write = cfi_amdext_write_2_by_16;
+			break;
+
+		default:
+		  printk("Unsupported interleave\n");
+			kfree(mtd);
+	  	kfree(cfi->cmdset_priv);
+		  return NULL;
+		}
+    break;
+
+	default:
+	  printk("Unsupported buswidth\n");
+		kfree(mtd);
+	  kfree(cfi->cmdset_priv);
+	  return NULL;
+    break;
+	}
 	mtd->sync = cfi_amdext_sync;
 	mtd->suspend = cfi_amdext_suspend;
 	mtd->resume = cfi_amdext_resume;
@@ -219,9 +272,9 @@
 
 	adr += chip->start;
 
-	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4);
-	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4);
-	map->write32(map, cpu_to_le32(0x00A000A0), 0x555 *4);
+	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00A000A0), 0x555 *4 + chip->start);
 	map->write32(map, cpu_to_le32(datum), adr);
 
 	spin_unlock_bh(chip->mutex);
@@ -346,11 +399,11 @@
 
 	adr += chip->start;
 
-	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4);
-	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4);
-	map->write32(map, cpu_to_le32(0x00800080), 0x555 *4);
-	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4);
-	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4);
+	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00800080), 0x555 *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00AA00AA), 0x555 *4 + chip->start);
+	map->write32(map, cpu_to_le32(0x00550055), 0x2AA *4 + chip->start);
 	map->write32(map, cpu_to_le32(0x00300030), adr);
 
 	
@@ -363,7 +416,7 @@
 	/* FIXME. Use a timer to check this, and return immediately. */
 	/* Once the state machine's known to be working I'll do that */
 
-	while ( ( (status = le32_to_cpu(map->read32(map, 0x00))) & 0x80808080 ) != 0x80808080 ) {
+	while ( ( (status = le32_to_cpu(map->read32(map, adr))) & 0x80808080 ) != 0x80808080 ) {
 		static int z=0;
 
 		if (chip->state != FL_ERASING) {
@@ -466,6 +519,400 @@
 
 
 
+static inline int do_read_2_by_8_onechip(struct map_info *map, struct flchip *chip, loff_t adr, size_t len, u_char *buf)
+{
+	DECLARE_WAITQUEUE(wait, current);
+	unsigned long timeo = jiffies + HZ;
+
+ retry:
+	spin_lock_bh(chip->mutex);
+
+	if (chip->state != FL_READY){
+printk("Waiting for chip to read, status = %d\n", chip->state);
+		set_current_state(TASK_INTERRUPTIBLE);
+		add_wait_queue(&chip->wq, &wait);
+
+		spin_unlock_bh(chip->mutex);
+
+		schedule();
+		remove_wait_queue(&chip->wq, &wait);
+
+		if(signal_pending(current))
+			return -EINTR;
+
+		timeo = jiffies + HZ;
+
+		goto retry;
+	}	
+
+	adr += chip->start;
+
+//     	map->write32(map, cpu_to_le32(0x00F000F0), adr);
+
+	chip->state = FL_READY;
+
+	map->copy_from(map, buf, adr, len);
+
+	wake_up(&chip->wq);
+	spin_unlock_bh(chip->mutex);
+
+	return 0;
+}
+
+static int cfi_amdext_read_2_by_8 (struct mtd_info *mtd, loff_t from, size_t len, size_t *retlen, u_char *buf)
+{
+	struct map_info *map = mtd->priv;
+	struct cfi_private *cfi = map->fldrv_priv;
+	unsigned long ofs;
+	int chipnum;
+	int ret = 0;
+
+	/* ofs: offset within the first chip that the first read should start */
+
+	chipnum = (from >> cfi->chipshift);
+	chipnum /= (cfi->interleave);
+	ofs = from - (chipnum <<  cfi->chipshift) * (cfi->interleave);
+
+	*retlen = 0;
+
+	while (len) {
+		unsigned long thislen;
+
+		if (chipnum >= cfi->numchips)
+			break;
+
+		if (((len + ofs -1) >> cfi->chipshift) / (cfi->interleave))
+			thislen = (1<<cfi->chipshift) * (cfi->interleave) - ofs;
+		else
+			thislen = len;
+
+		ret = do_read_2_by_8_onechip(map, &cfi->chips[chipnum], ofs, thislen, buf);
+		if (ret)
+			break;
+
+		*retlen += thislen;
+		len -= thislen;
+		buf += thislen;
+
+		ofs = 0;
+		chipnum++;
+	}
+	return ret;
+}
+
+static inline int do_write_2_by_8_oneword(struct map_info *map, struct flchip *chip, unsigned long adr, __u16 datum)
+{
+	unsigned long timeo = jiffies + HZ;
+	unsigned int Last[4];
+	unsigned long Count = 0;
+	DECLARE_WAITQUEUE(wait, current);
+	int ret = 0;
+
+ retry:
+	spin_lock_bh(chip->mutex);
+
+	if (chip->state != FL_READY){
+printk("Waiting for chip to write, status = %d\n", chip->state);
+		set_current_state(TASK_INTERRUPTIBLE);
+		add_wait_queue(&chip->wq, &wait);
+
+		spin_unlock_bh(chip->mutex);
+
+		schedule();
+		remove_wait_queue(&chip->wq, &wait);
+printk("Wake up to write:\n");
+		if(signal_pending(current))
+			return -EINTR;
+
+		timeo = jiffies + HZ;
+
+		goto retry;
+	}	
+
+	chip->state = FL_WRITING;
+
+	adr += chip->start;
+
+	map->write16(map, cpu_to_le16(0xAAAA), 0x555 *2 + chip->start);
+	map->write16(map, cpu_to_le16(0x5555), 0x2AA *2 + chip->start);
+	map->write16(map, cpu_to_le16(0xA0A0), 0x555 *2 + chip->start);
+	map->write16(map, cpu_to_le16(datum), adr);
+
+	spin_unlock_bh(chip->mutex);
+	udelay(chip->word_write_time);
+	spin_lock_bh(chip->mutex);
+
+	Last[0] = map->read16(map, adr);
+	Last[1] = map->read16(map, adr);
+	Last[2] = map->read16(map, adr);
+
+	for (Count = 3; Last[(Count - 1) % 4] != Last[(Count - 2) % 4] && Count < 10000; Count++){
+		udelay(10);
+
+		Last[Count % 4] = map->read16(map, adr);
+	}
+
+	if (Last[(Count - 1) % 4] != datum){
+	     	map->write16(map, cpu_to_le16(0xF0F0), adr);
+
+		ret = -EIO;
+	}
+
+	chip->state = FL_READY;
+	wake_up(&chip->wq);
+	spin_unlock_bh(chip->mutex);
+
+	return ret;
+}
+
+
+static int cfi_amdext_write_2_by_8 (struct mtd_info *mtd, loff_t to , size_t len, size_t *retlen, const u_char *buf)
+{
+	struct map_info *map = mtd->priv;
+	struct cfi_private *cfi = map->fldrv_priv;
+	int ret = 0;
+	int chipnum;
+	unsigned long ofs;
+
+	*retlen = 0;
+
+	chipnum = (to >> cfi->chipshift);
+	chipnum /= cfi->interleave;
+	ofs = to  - (chipnum << cfi->chipshift) * cfi->interleave;
+
+  if (ofs & 1) {
+		/* destination doesn't start on word boundary, so write the first odd byte */
+  	unsigned int tmp;
+
+#if defined(__LITTLE_ENDIAN)
+		/* get existing low byte of word */
+		tmp = map->read8(map, (ofs & ~1) + cfi->chips[chipnum].start);
+
+		/* add in the first byte we're writing */
+		tmp |= (*buf << 8);
+
+		ret = do_write_2_by_8_oneword(map, &cfi->chips[chipnum], ofs, tmp);
+
+#elif defined(__BIG_ENDIAN)
+#error not support big endian yet
+#else
+#error define a sensible endianness
+#endif
+
+	  if (ret)
+			return ret;
+
+		ofs += 1;
+		buf += 1;
+		(*retlen) += 1;
+		len -= 1;
+
+		if ((ofs >> cfi->chipshift) / cfi->interleave) {
+			chipnum ++;
+			ofs = 0;
+			if (chipnum == cfi->numchips)
+				return 0;
+		}
+  }
+
+	while(len > 1) {
+		ret = do_write_2_by_8_oneword(map, &cfi->chips[chipnum],
+					       ofs, *(__u16 *)buf);
+		if (ret)
+			return ret;
+
+		ofs += 2;
+		buf += 2;
+		(*retlen) += 2;
+		len -= 2;
+
+		if ((ofs >> cfi->chipshift) / cfi->interleave) {
+			chipnum ++;
+			ofs = 0;
+			if (chipnum == cfi->numchips)
+				return 0;
+		}
+	}
+
+	if (len) {
+		unsigned int tmp;
+
+		/* Final byte to write */
+#if defined(__LITTLE_ENDIAN)
+		/* get existing high byte of word */
+		tmp = map->read16(map, ofs + cfi->chips[chipnum].start) & 0xff00;
+
+		/* add in last byte */
+		tmp |= *buf;
+
+		ret = do_write_2_by_8_oneword(map, &cfi->chips[chipnum], ofs, tmp);
+
+#elif defined(__BIG_ENDIAN)
+#error not support big endian yet
+#else
+#error define a sensible endianness
+#endif
+
+		if (ret)
+			return ret;
+		
+		(*retlen) += len;
+	}
+
+	return 0;
+}
+
+
+static inline int do_erase_2_by_8_oneblock(struct map_info *map, struct flchip *chip, unsigned long adr)
+{
+	unsigned int status;
+	unsigned long timeo = jiffies + HZ;
+	DECLARE_WAITQUEUE(wait, current);
+
+ retry:
+	spin_lock_bh(chip->mutex);
+
+	if (chip->state != FL_READY){
+		set_current_state(TASK_INTERRUPTIBLE);
+		add_wait_queue(&chip->wq, &wait);
+
+		spin_unlock_bh(chip->mutex);
+
+		schedule();
+		remove_wait_queue(&chip->wq, &wait);
+
+		if(signal_pending(current))
+			return -EINTR;
+
+		timeo = jiffies + HZ;
+
+		goto retry;
+	}	
+
+	chip->state = FL_ERASING;
+
+	adr += chip->start;
+
+	map->write16(map, cpu_to_le16(0xAAAA), 0x555 *2 + chip->start);
+	map->write16(map, cpu_to_le16(0x5555), 0x2AA *2 + chip->start);
+	map->write16(map, cpu_to_le16(0x8080), 0x555 *2 + chip->start);
+	map->write16(map, cpu_to_le16(0xAAAA), 0x555 *2 + chip->start);
+	map->write16(map, cpu_to_le16(0x5555), 0x2AA *2 + chip->start);
+	map->write16(map, cpu_to_le16(0x3030), adr);
+
+	
+	timeo = jiffies + (HZ*20);
+
+	spin_unlock_bh(chip->mutex);
+	schedule_timeout(HZ);
+	spin_lock_bh(chip->mutex);
+
+	/* FIXME. Use a timer to check this, and return immediately. */
+	/* Once the state machine's known to be working I'll do that */
+
+	while ( ( (status = le16_to_cpu(map->read16(map, adr))) & 0x8080 ) != 0x8080 ) {
+		static int z=0;
+
+		if (chip->state != FL_ERASING) {
+			/* Someone's suspended the erase. Sleep */
+			set_current_state(TASK_INTERRUPTIBLE);
+			add_wait_queue(&chip->wq, &wait);
+			
+			spin_unlock_bh(chip->mutex);
+			printk("erase suspended. Sleeping\n");
+			
+			schedule();
+			remove_wait_queue(&chip->wq, &wait);
+			
+			if (signal_pending(current))
+				return -EINTR;
+			
+			timeo = jiffies + (HZ*2); /* FIXME */
+			spin_lock_bh(chip->mutex);
+			continue;
+		}
+
+		/* OK Still waiting */
+		if (time_after(jiffies, timeo)) {
+			chip->state = FL_READY;
+			spin_unlock_bh(chip->mutex);
+			printk("waiting for erase to complete timed out.");
+			return -EIO;
+		}
+		
+		/* Latency issues. Drop the lock, wait a while and retry */
+		spin_unlock_bh(chip->mutex);
+
+		z++;
+		if ( 0 && !(z % 100 ))
+			printk("chip not ready yet after erase. looping\n");
+
+		udelay(1);
+		
+		spin_lock_bh(chip->mutex);
+		continue;
+	}
+	
+	/* Done and happy. */
+	chip->state = FL_READY;
+	wake_up(&chip->wq);
+	spin_unlock_bh(chip->mutex);
+	printk("erase ret OK\n");
+	return 0;
+}
+
+static int cfi_amdext_erase_2_by_8 (struct mtd_info *mtd, struct erase_info *instr)
+{
+	struct map_info *map = mtd->priv;
+	struct cfi_private *cfi = map->fldrv_priv;
+	unsigned long adr, len;
+	int chipnum, ret = 0;
+
+//printk("erase : 0x%x 0x%x 0x%x\n", instr->addr, instr->len, mtd->size);
+
+	if (instr->addr & (mtd->erasesize - 1))
+		return -EINVAL;
+
+	if (instr->len & (mtd->erasesize -1))
+		return -EINVAL;
+
+	if ((instr->len + instr->addr) > mtd->size)
+		return -EINVAL;
+
+	chipnum = instr->addr >> cfi->chipshift;
+	chipnum /= cfi->interleave;
+	adr = instr->addr - (chipnum << cfi->chipshift) * (cfi->interleave);
+	len = instr->len;
+
+printk("erase : 0x%x 0x%x 0x%x\n", adr, len, chipnum, mtd->size);
+
+	while(len) {
+//printk("erase : 0x%x 0x%x 0x%x 0x%x\n", chipnum, adr, len, cfi->chipshift);
+		ret = do_erase_2_by_8_oneblock(map, &cfi->chips[chipnum], adr);
+
+		if (ret)
+			return ret;
+
+		adr += mtd->erasesize;
+		len -= mtd->erasesize;
+
+		if ((adr >> cfi->chipshift) / (cfi->interleave)) {
+			adr = 0;
+			chipnum++;
+			
+			if (chipnum >= cfi->numchips)
+			break;
+		}
+	}
+		
+	if (instr->callback)
+		instr->callback(instr);
+	
+	return 0;
+}
+
+
+
 static void cfi_amdext_sync (struct mtd_info *mtd)
 {
 	struct map_info *map = mtd->priv;
@@ -474,7 +921,6 @@
 	struct flchip *chip;
 	int ret = 0;
 	DECLARE_WAITQUEUE(wait, current);
-printk("sync\n");
 
 	for (i=0; !ret && i<cfi->numchips; i++) {
 		chip = &cfi->chips[i];
@@ -493,6 +939,7 @@
 			 * as the whole point is that nobody can do anything
 			 * with the chip now anyway.
 			 */
+		case FL_SYNCING:
 			spin_unlock_bh(chip->mutex);
 			break;
 
@@ -520,7 +967,6 @@
 		}
 		spin_unlock_bh(chip->mutex);
 	}
-printk("sync end\n");
 }
 
 
@@ -549,6 +995,7 @@
 			 * as the whole point is that nobody can do anything
 			 * with the chip now anyway.
 			 */
+		case FL_PM_SUSPENDED:
 			spin_unlock_bh(chip->mutex);
 			break;
 
